"""
computer_graphics_demo.reality
==============================

Draw a scene by hidden surface removal, mirror reflection, texture, etc.
"""
import taichi as ti
import numpy as np
import math
import random
import itertools
from . import pixels

# 0: x0, 1: y0, 2: z0, 3: r, 4: type, 5: R, 6: G, 7: B
# type = 0: diffuse reflection
# type = 1: mirror reflection
# type = 2: refraction, 5: refreaction rate
spheres = ti.Vector.field(8, float)


def init():
    """init."""
    # earth
    spheres[0] = np.array([0.0, -1000.0, 0.0, 1000.0, 0.0, 0.5, 0.5, 0.5])
    # diffuse reflection example
    spheres[1] = np.array([-4.0, 1.0, 0.0, 1.0, 0.0, 0.4, 0.2, 0.1])
    # refraction example
    # you will see a reversal graph through a transparent ball
    # https://img-blog.csdn.net/20170114205419889?watermark/2/text/aHR0cDovL2Jsb2cuY3Nkbi5uZXQvbGliaW5nX3plbmc=/font/5a6L5L2T/fontsize/400/fill/I0JBQkFCMA==/dissolve/70/gravity/Center
    spheres[2] = np.array([4.0, 1.0, 0.0, 1.0, 2.0, 1.5, 0.0, 0.0])
    # mirror reflection example
    spheres[3] = np.array([0.0, 1.0, 0.0, 1.0, 1.0, 0.7, 0.6, 0.5])
    k, y, ran = 4, 0.5, range(-10, 10, 6)
    for x, z in itertools.product(ran, ran):
        refract_type = random.randint(0, 2)
        spheres[k] = np.array(
            [
                x,
                y,
                z,
                y,
                refract_type,
                random.random() if refract_type == 2 else random.random() + 1,
                random.random(),
                random.random(),
            ],
        )
        k += 1


nx = 1024
ny = 512
lookfrom = np.array([6.5, 1.5, 2.0])
lookat = np.array([0.0, 0.0, 0.0])
up = np.array([0.0, 1.0, 0.0])
view_dir = lookfrom - lookat
dist_to_focus = np.sqrt((view_dir * view_dir).sum())  # type: ignore


def normalize(n):
    """normalize.

    :param n:
    """
    return n / math.sqrt(np.dot(n, n))


# u, v, w is unit vector of relative coordinate
# u is horizontal
# v is vertical
w = normalize(view_dir)
u = normalize(np.cross(up, w))
v = np.cross(w, u)
# make sure the watching scope is suit
horizontal = u * 2 * dist_to_focus
vertical = v * 2 * ny / nx * dist_to_focus


@ti.func
def unit_vector(v):
    """unit_vector.

    :param v:
    """
    k = 1 / (ti.sqrt(v.dot(v)))
    return k * v


@ti.func
def reflect(v, n):
    """reflect.

    https://en.wikipedia.org/wiki/Specular_reflection

    :param v:
    :param n:
    """
    return v - 2 * v.dot(n) * n


@ti.func
def refract(v, n, ni_over_nt):
    """refract.

    https://en.wikipedia.org/wiki/Snell%27s_law

    :param v:
    :param n:
    :param ni_over_nt:
    """
    dt = v.dot(n)
    discriminant = 1.0 - ni_over_nt * ni_over_nt * (1.0 - dt * dt)
    succ = False
    refracted = ti.Vector([0.0, 0.0, 0.0])
    if discriminant > 0:
        refracted = ni_over_nt * (v - n * dt) - n * ti.sqrt(discriminant)
        succ = True
    return succ, refracted


@ti.func
def schlick(cosine, ni_over_nt):
    """schlick.

    https://en.wikipedia.org/wiki/Schlick%27s_approximation

    :param cosine:
    :param ref_idx:
    """
    # cosine must > 0, or result will > 1
    # however, result is reflection coefficient
    # i.e., ratio of the amplitude of the reflected wave to the incident wave
    # must < 1
    # wikipedia is not fully correct.
    cosine = ti.abs(cosine)
    r0 = ((1 - ni_over_nt) / (1 + ni_over_nt)) ** 2
    return r0 + (1 - r0) * ti.pow((1 - cosine), 5)


@ti.func
def random_in_unit_sphere():
    """random_in_unit_sphere.

    http://www.pbr-book.org/3ed-2018/Monte_Carlo_Integration/2D_Sampling_with_Multidimensional_Transformations.html
    """
    eta1, eta2 = ti.random(), ti.random()
    coeff = 2 * ti.sqrt(eta1 * (1 - eta1))
    eta2m2pi = eta2 * math.pi * 2
    return ti.Vector(
        [ti.cos(eta2m2pi) * coeff, ti.sin(eta2m2pi) * coeff, 1 - 2 * eta1]
    )


@ti.func
def hit_sphere(center, radius, rayo, rayd, mint, maxt):
    """hit_sphere.

    :param center:
    :param radius:
    :param rayo:
    :param rayd:
    :param mint:
    :param maxt:
    """
    rst = False
    t = 0.0
    oc = rayo - center
    a = rayd.dot(rayd)
    b = 2.0 * oc.dot(rayd)
    c = oc.dot(oc) - radius * radius
    disc = b * b - 4.0 * a * c
    if disc >= 0:
        sq = ti.sqrt(disc)
        t = (-b - sq) / (2.0 * a)
        if t > mint and t < maxt:
            rst = True
        else:
            t = (-b + sq) / (2.0 * a)
            if t > mint and t < maxt:
                rst = True

    return rst, t, rayo + t * rayd


@ti.func
def hit_spheres(rayo, rayd, mint, maxt):
    """hit_spheres.

    :param rayo:
    :param rayd:
    :param mint:
    :param maxt:
    """
    rst = False
    rt = 1e10
    hit_point = ti.Vector([0.0, 0.0, 0.0])
    normal = ti.Vector([0.0, 0.0, 0.0])
    index = -1
    for i in ti.static(range(20)):
        sc = ti.Vector([spheres[i][0], spheres[i][1], spheres[i][2]])
        r = spheres[i][3]
        hit, t, hitp = hit_sphere(sc, r, rayo, rayd, mint, maxt)
        if hit and t < rt:
            rst = True
            rt = t
            normal = unit_vector(hitp - sc)
            hit_point = hitp
            index = i
    return rst, hit_point, normal, index


@ti.func
def color(start, dir):
    """color.

    :param start:
    :param dir:
    """
    rst = ti.Vector([1.0, 1.0, 1.0])
    count = 0
    while count < 10:
        hit, hit_point, n, index = hit_spheres(start, dir, 0.001, 1e10)
        if hit:
            # hit_point is next start point of ray
            start = hit_point
            if spheres[index][4] <= 1.0:
                albedo = ti.Vector(
                    [spheres[index][5], spheres[index][6], spheres[index][7]]
                )
                rst = rst * albedo
                if spheres[index][4] == 0.0:
                    dir = unit_vector(n + random_in_unit_sphere())
                elif spheres[index][4] == 1.0:
                    dir = reflect(dir, n)
            else:
                cosine = dir.dot(n)
                # ray comes from outer to inner
                # outward_n in the semi-space where incident vector exists
                outward_n = n
                ni_over_nt = 1.0 / spheres[index][5]
                # ray comes from inner to outer
                if cosine > 0:
                    outward_n = -n
                    ni_over_nt = spheres[index][5]
                succ, refracted = refract(dir, outward_n, ni_over_nt)
                if succ:
                    reflect_prob = schlick(cosine, ni_over_nt)
                    # reflect_prob is < 10%, so refraction is major
                    # reflection only make the ray dim
                    if ti.random() < reflect_prob:
                        dir = reflect(dir, n)
                    else:
                        dir = refracted
                # full reflect
                else:
                    dir = reflect(dir, n)
            count += 1
        else:
            t = 0.5 * (dir[1] + 1.0)
            skycolor = (1.0 - t) * ti.Vector([1.0, 1.0, 1.0]) + t * ti.Vector(
                [0.5, 0.7, 1.0]
            )
            rst = rst * skycolor
            break
    return rst


@ti.kernel
def paint(R: int, G: int, B: int, sample_number: int, r_aperture: float):
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param sample_number:
    :type sample_number: int
    :param r_aperture:
    :type r_aperture: float
    """
    for i, j in pixels:
        col = ti.Vector([0.0, 0.0, 0.0])
        for _ in range(sample_number):
            # change vector of uOv plane to Oxyz
            uu = float(i) / float(nx)
            vv = float(j) / float(ny)
            uv = ti.Vector([uu, vv])
            # use a random offset to reduce sawtooth
            # use another random distribution is possible
            # e.g., vector norm is ti.random()
            # and theta is ti.random() * 2 * math.pi
            offset = r_aperture * ti.Vector([ti.random(), ti.random()]).dot(uv)
            direction = (
                uu * ti.Vector(horizontal)
                + vv * ti.Vector(vertical)
                - ti.Vector(horizontal + vertical) / 2
                - ti.Vector(lookfrom)
                - offset
            )
            col += color(ti.Vector(lookfrom) + offset, unit_vector(direction))
        pixels[i, j] = col / sample_number



"""
computer_graphics_demo.fill.scan
================================

Scan line algorithm.
"""
import taichi as ti
from typing import List, Tuple, Optional
from .. import setpixel
from . import cycle_shift


def paint(R: int, G: int, B: int, *argv: int):
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param argv:
    :type argv: int
    """
    if len(argv) % 2:
        exit(f"point number must be even, now is {len(argv)}")
    elif len(argv) < 6:
        exit(f"point number must be greater than 3, now is {len(argv)}")
    xs = argv[::2]
    ys = argv[1::2]
    vertexs = list(zip(xs, ys))
    y_max = max(ys)
    vertexs0 = cycle_shift(vertexs, -1)
    vertexs2 = cycle_shift(vertexs, 1)
    vertexs3 = cycle_shift(vertexs2, 1)
    edges = zip(vertexs0, vertexs, vertexs2, vertexs3)  # type: ignore
    edge_table: List[Optional[List[Tuple[int, float, float]]]] = [None] * y_max

    # sorted edge table
    for (_, y0), (x1, y1), (x2, y2), (_, y3) in edges:
        if y1 == y2:
            continue
        dx = (x2 - x1) / (y2 - y1)
        ymin, ymax = ti.min(y1, y2), ti.max(y1, y2)
        if ymin == y1:
            xstart = x1 + 0.0
        else:
            xstart = x2 + 0.0
        # https://github.com/jingedawang/PolygonScan/blob/master/main.cpp#L109
        if y0 < y1 < y2 or y1 > y2 > y3:
            ymin += 1
            xstart += dx
        if edge_table[ymin]:
            edge_table[ymin] += [(ymax, xstart, dx)]  # type: ignore
        else:
            edge_table[ymin] = [(ymax, xstart, dx)]

    # active edge table
    edge_list_old = None
    for y in range(y_max):
        if edge_list_old:
            for ymax, xstart, dx in edge_list_old:
                if y <= ymax:
                    if edge_table[y]:
                        edge_table[y] += [(ymax, xstart + dx, dx)]
                        edge_table[y] = sorted(
                            edge_table[y], key=lambda x: x[2]  # type: ignore
                        )
                        edge_table[y] = sorted(
                            edge_table[y], key=lambda x: x[1]  # type: ignore
                        )
                    else:
                        edge_table[y] = [(ymax, xstart + dx, dx)]
        if edge_table[y]:
            edge_list_old = edge_table[y]
            for i in range(len(edge_table[y])):  # type: ignore
                if i % 2 == 1:
                    continue
                x_start = ti.ceil(edge_table[y][i][1])  # type: ignore
                x_end = ti.floor(edge_table[y][i + 1][1]) + 1  # type: ignore
                for x in range(x_start, x_end):
                    setpixel(x, y, R, G, B)



"""
computer_graphics_demo.b
========================

Draw a B-spline.
"""
import taichi as ti
from .line import polygon
from .line.bresenham import line
from typing import List

nodes = [0, 0, 0, 0, 100, 200, 300, 400, 500, 500, 500, 500]


def b_basis(i: int, degree: int, u: int, nodes: List[int]) -> float:
    """b_basis.

    :param i:
    :type i: int
    :param degree:
    :type degree: int
    :param u:
    :type u: int
    :param nodes:
    :type nodes: List[int]
    :rtype: float
    """
    if degree:
        length1 = nodes[i + degree] - nodes[i]
        length2 = nodes[i + degree + 1] - nodes[i + 1]
        alpha = (u - nodes[i]) / length1 if length1 else 0
        beta = (nodes[i + degree + 1] - u) / length2 if length2 else 0
        result = alpha * b_basis(i, degree - 1, u, nodes) + beta * b_basis(
            i + 1, degree - 1, u, nodes
        )
    else:
        result = float(nodes[i] < u <= nodes[i + 1])
    return result


def paint(R: int, G: int, B: int, degree: int, *argv: int) -> None:
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param degree:
    :type degree: int
    :param argv:
    :type argv: int
    :rtype: None
    """
    xs = argv[::2]
    ys = argv[1::2]
    if len(argv) % 2:
        exit(f"point number must be even, now is {len(argv)}")
    elif len(argv) < 6:
        n_node = len(xs) + degree + 1
        exit(f"node number must be greater than {n_node}, now is {len(nodes)}")
    argv2 = []
    for u in range(nodes[0], nodes[-1]):
        x, y = 0, 0
        for i in range(len(xs)):
            x += xs[i] * b_basis(i, degree, u, nodes)
            y += ys[i] * b_basis(i, degree, u, nodes)
        argv2 += [ti.floor(x), ti.floor(y)]
    polygon(R, G, B, *argv2, line=line)



"""
computer_graphics_demo.bezier
=============================

Draw a Bezier curve.
"""
import taichi as ti
from .line import polygon
from .line.bresenham import line


def basis(i: int, degree: int, u: float) -> float:
    """basis.

    :param i:
    :type i: int
    :param degree:
    :type degree: int
    :param u:
    :type u: float
    :rtype: float
    """
    if i > degree:
        return 0
    elif i == 0:
        return (1 - u) ** degree
    elif i == 1:
        return degree * u * (1 - u) ** (degree - 1)
    return basis(degree - 1, i, u) * (1 - u) + basis(degree - 1, i - 1, u) * u


def paint(R: int, G: int, B: int, sample_number: int, *argv: int) -> None:
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param sample_number:
    :type sample_number: int
    :param argv:
    :type argv: int
    :rtype: None
    """
    xs = argv[::2]
    ys = argv[1::2]
    if len(argv) % 2:
        exit(f"point number must be even, now is {len(argv)}")
    argv2 = []
    for u in range(0, sample_number):
        x, y = 0, 0
        for i in range(len(xs)):
            x += xs[i] * basis(i, len(xs) - 1, u / sample_number)
            y += ys[i] * basis(i, len(xs) - 1, u / sample_number)
        argv2 += [ti.floor(x), ti.floor(y)]
    polygon(R, G, B, *argv2, line=line)



# noqa: D400 D205
"""
computer_graphics_demo
======================

Common functions.
"""
import logging
from typing import Final

import taichi as ti

logger = logging.getLogger(__name__)

try:
    from get_version import get_version, NoVersionFound

    try:
        __version__ = get_version(__file__)
    except NoVersionFound:
        __version__ = "0.0.0"
except ImportError:
    logger.warning("Please install get_version!")
    __version__ = "0.0.0"

_binname: Final = "cgdemo"
pixels = ti.Vector.field(3, float)


def init() -> None:
    """init.

    :rtype: None
    """
    pass


@ti.pyfunc
def setpixel(x: int, y: int, R: int, G: int, B: int):
    """setpixel.

    :param x:
    :type x: int
    :param y:
    :type y: int
    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    """
    pixels[x, y] = ti.Vector([R, G, B])



"""
computer_graphics_demo.koch
===========================

Draw a Koch curve
"""
import math
from .line.bresenham import line


def rule(char: str) -> str:
    """rule.

    :param char:
    :type char: str
    :rtype: str
    """
    return "F+F--F+F" if char == "F" else char


def paint(R: int, G: int, B: int, scale: int, iteration_max: int, x0: int,
          y0: int):
    """paint.

    :param R:
    :type R: int
    :param G:
    :type G: int
    :param B:
    :type B: int
    :param scale:
    :type scale: int
    :param iteration_max:
    :type iteration_max: int
    :param x0:
    :type x0: int
    :param y0:
    :type y0: int
    """
    info = "F"
    for _ in range(iteration_max):
        info = "".join(map(rule, info))
    x1, y1, angle = x0, y0, 0
    for char in info:
        if char == "F":
            x2 = x1 + scale * math.cos(angle)
            y2 = y1 + scale * math.sin(angle)
            line(R, G, B, math.floor(x1), math.floor(y1),
                 math.floor(x2), math.floor(y2))
            x1, y1 = x2, y2
        elif char == "+":
            angle += math.pi / 3
        elif char == "-":
            angle -= math.pi / 3



"""
computer_graphics_demo.fill
===========================

Fill a region.
"""
import copy


def cycle_shift(xs, number: int):
    """cycle_shift.

    :param xs:
    :param number:
    :type number: int
    """
    ys = copy.copy(xs)
    num = abs(number)
    for _ in range(num):
        if number < 0:
            ys.insert(0, ys.pop())
        else:
            ys.insert(len(ys), ys[0])
            ys.remove(ys[0])
    return ys
